//
//  RTcmixPlayer.h
//  subclass this singleton for iRTcmix audio work
//	will function in both AudioQueueBuffer and AudioUnit mode
//
//  Created by Brad Garton on 7/17/09.
//	additional work by Damon Holzborn
//

#include <AudioToolbox/AudioToolbox.h>
#import "RTcmixPlayer.h"

static RTcmixPlayer *sharedSingletonManager = nil;

@implementation RTcmixPlayer

// Audio Queue
@synthesize bufferByteSize;
@synthesize audioPlayerShouldStopImmediately;
@synthesize queueObject;
@synthesize audioFormat;

// AudioUnit
@synthesize inputProc;
@synthesize rioUnit;
@synthesize outputBufferIndex;
@synthesize rtcmixOutputBuffer;
@synthesize inputBufferIndex;
@synthesize rtcmixInputBuffer;
@synthesize audioUnitFlag;
@synthesize audioInputFlag;

#pragma mark Data and Communication

// send score to RTcmix with initialized parameters
- (void) parseScoreWithRTcmixScore:(RTcmixScore *)score {
	NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
	RTcmixScore *scoreObject = score;
	
	if (![scoreObject setupIsRunning] && [scoreObject setupScore] != nil) 
	{
		NSString *setupScore = [NSString stringWithString:[scoreObject setupScore]];
		NSArray *setupScoreParameters = [NSArray arrayWithArray:[scoreObject setupScoreParameters]];
		char setupCScore[16384];
		char *setupScoreArgList = (char *)malloc(sizeof(NSString *) * [setupScoreParameters count]);
		[setupScoreParameters getObjects:(id *)setupScoreArgList];
		NSString *setupScoreFormatted = [[[NSString alloc] initWithFormat:setupScore arguments:setupScoreArgList] autorelease];
		strcpy (setupCScore,[setupScoreFormatted  UTF8String]);
		parse_score(setupCScore, strlen(setupCScore));
		
		free(setupScoreArgList);
	}
	
	NSString *mainScore = [NSString stringWithString:[scoreObject mainScore]];
	NSArray *mainScoreParameters = [NSArray arrayWithArray:[scoreObject mainScoreParameters]];
	char mainCScore[16384];
	
	char *mainScoreArgList = (char *)malloc(sizeof(NSString *) * [mainScoreParameters count]);
    [mainScoreParameters getObjects:(id *)mainScoreArgList];
	NSString *mainScoreFormatted = [[[NSString alloc] initWithFormat:mainScore arguments:mainScoreArgList] autorelease];
	strcpy (mainCScore,[mainScoreFormatted UTF8String]);
	parse_score(mainCScore, strlen(mainCScore));
	
	free(mainScoreArgList);
    [pool release];
}

- (void) parseScoreWithNSString:(NSString *)score {
	char cScore[16384];
	strcpy (cScore,[score UTF8String]);
	parse_score(cScore, strlen(cScore));
}

// use this to get data in via the makeconnection() PField scheme
- (void) setInlet:(int)inlet withValue:(Float32)value {
	pfield_set(inlet, value);	
}

// maxBang and maxMessage is called from within the playbackCallback(), 
// set by MAXBANG() and MAXMESSAGE() in the score - override to implement
- (void)maxBang { } 

- (void)maxMessage { };

// stop all scheduled scripts
- (void)flushAllScripts {
	flush_sched();
}

// soundfileplayback from memory
- (int)setSampleBuffer:(NSString *)bufferName withSoundFile:(NSString *)soundFileName {
	AudioFileID     soundFile;
	AudioStreamBasicDescription		dataFormat;
	OSStatus status;
	UInt32 propsize;

	void *sampleBuffer;
	float *floatSampleBuffer;

	NSString *soundFilePath = [[NSBundle mainBundle]pathForResource:soundFileName ofType:NULL];
	NSURL *soundFileURL = [NSURL fileURLWithPath:soundFilePath];

	// BGG -- should probably use ExtAudioFileOpenURL()
	status = AudioFileOpenURL((CFURLRef)soundFileURL, kAudioFileReadPermission, kAudioFileAIFFType, &soundFile);
	if (status != 0) 
	{
		NSLog(@"error opening soundfile");
		return 0;
	}

	propsize = sizeof(dataFormat);
	status = AudioFileGetProperty(soundFile, kAudioFilePropertyDataFormat, &propsize, &dataFormat);
	if (status != 0) 
	{
		NSLog(@"problem geting AudioFile properties");
		return 0;
	}

	//Get file size
	UInt64 outDataSize = 0;
	propsize = sizeof(UInt64);
	AudioFileGetProperty(soundFile, kAudioFilePropertyAudioDataByteCount,
		&propsize, &outDataSize);

	//Get packet count
	UInt64 outPackets = 0;
	propsize = sizeof(UInt64);
	AudioFileGetProperty(soundFile, kAudioFilePropertyAudioDataPacketCount,
		&propsize, &outPackets);

	UInt32 numBytes = outDataSize;
	UInt32 numPackets = outPackets;
	sampleBuffer = malloc(outDataSize);
	if (sampleBuffer == NULL) 
	{
		NSLog(@"error mallocing sample buffer");
		return 0;
	}

	status = AudioFileReadPackets (soundFile, false, &numBytes, NULL, 0, &numPackets, sampleBuffer);
	if (status !=0) 
	{
		NSLog(@"problem reading file packets into buffer");
		return 0;
	}

	floatSampleBuffer = malloc(numPackets*dataFormat.mChannelsPerFrame*sizeof(float));
	if (floatSampleBuffer == NULL) 
	{
		NSLog(@"error mallocing fsample buffer");
		return 0;
	}

	// convert to floating-point buffer for RTcmix
	float *fsbptr = floatSampleBuffer;
	short *sbptr = sampleBuffer;
	for(int i = 0; i < (numPackets*dataFormat.mChannelsPerFrame); i++) {
		sbptr[i] = CFSwapInt16(sbptr[i]);
		fsbptr[i] = (float)sbptr[i]/32768.0;
	}

	buffer_set((char *)[bufferName UTF8String], fsbptr, numPackets, dataFormat.mChannelsPerFrame, 0);
	
	return 1;
}

#pragma mark Audio

- (void)startAudioQueuePlayer {
	[self setupAudioSettings];
	
	// initialize RTcmix
	rtcmixmain();
	
	// before instantiating the playback audio queue object, 
	//	set the audio session category
	UInt32 sessionCategory = kAudioSessionCategory_MediaPlayback;
	AudioSessionSetProperty (kAudioSessionProperty_AudioCategory,
							 sizeof (sessionCategory), &sessionCategory);		
	
	// initialize the audio session object for this application,
	//		registering the callback that Audio Session Services will invoke 
	//		when there's an interruption
	AudioSessionInitialize (NULL, NULL, interruptionListenerCallback, self);

	audioUnitFlag = 0; // not running as AU

	[self setupPlayback];
	[self setupPlaybackAudioQueueObject];
	[self setAudioPlayerShouldStopImmediately: NO];
	AudioSessionSetActive (true);
	
	[self playAudio];
}

- (void)startAUPlayerWithInputFlag:(int)inputFlag {
	[self setupAudioSettings];
	audioUnitFlag = 1; // running as AU
	audioInputFlag = inputFlag; // 1 if we are running input

	inputProc.inputProc = rtcmixPerformCallback;
	inputProc.inputProcRefCon = self;

	// initialize the audio session object for this application,
	//		registering the callback that Audio Session Services will invoke 
	//		when there's an interruption
	AudioSessionInitialize(NULL, NULL, interruptionListenerCallback, self);
	AudioSessionSetActive(true);

	//	set the audio session category
	UInt32 sessionCategory;
	if (audioInputFlag == 1)
		sessionCategory = kAudioSessionCategory_PlayAndRecord;
	else
		sessionCategory = kAudioSessionCategory_MediaPlayback;

	AudioSessionSetProperty(kAudioSessionProperty_AudioCategory,
		sizeof(sessionCategory), &sessionCategory);


	if (audioUnitFlag == 1) 
	{ //possibly do these for AudioQueue version too?
		// add a property listener
		AudioSessionAddPropertyListener(kAudioSessionProperty_AudioRouteChange,
			propListenerCallback, self);

		// this sets buffer size when using AudioUnit
		Float32 preferredBufferSize = .1;
		AudioSessionSetProperty(kAudioSessionProperty_PreferredHardwareIOBufferDuration,
			sizeof(preferredBufferSize), &preferredBufferSize);
	}

	Float64 prefSR = 44100.0;
	AudioSessionSetProperty(kAudioSessionProperty_PreferredHardwareSampleRate,
		sizeof(prefSR), &prefSR);

	// this is because input AudioUnits default to the phone/speaker output, and I want to hear it!
	UInt32 newRoute = kAudioSessionOverrideAudioRoute_Speaker;
	AudioSessionSetProperty(kAudioSessionProperty_OverrideAudioRoute,
		sizeof(newRoute), &newRoute);

	[self setupPlayback];
	[self setupRemoteIO];

	// initialize RTcmix
	rtcmixmain();
	// set rtcmix params (note we don't use inbuf and outbuf, they are passed in pulltraverse() above)
	maxmsp_rtsetparams(audioFormat.mSampleRate, audioFormat.mChannelsPerFrame, vectorSize, NULL, NULL, errbuf);

	outputBufferIndex = vectorSize*numberOfChannels; // set at max to get a pullTraverse() straightaway

	// set up the buffers used by RTcmix (asynch with RemoteIO
	rtcmixOutputBuffer = malloc(vectorSize*numberOfChannels*sizeof(short)); 
	if (rtcmixOutputBuffer == NULL) // ok, not much of a check...
		NSLog(@"rtcmixOutputBuffer is NULL!");

	inputBufferIndex = vectorSize*numberOfChannels;
	if (audioInputFlag == 1) 
	{
		rtcmixInputBuffer = malloc(vectorSize*numberOfChannels*sizeof(short));
		if (rtcmixInputBuffer == NULL)
			NSLog(@"rtcmixInputBuffer is NULL!");

		// zero out input buffer
		memset((void *)rtcmixInputBuffer, 0, vectorSize*numberOfChannels*sizeof(short));
	}

	[self setAudioPlayerShouldStopImmediately: NO];
	[self playAudio];
}

- (void)setupAudioSettings {	
	numberAudioDataBuffers = 7; // max 32
	vectorSize = 512; // increase this (in powers of 2) to improve performance (at the exense of latency)
	numberOfChannels = 2;
}

- (void) setupPlayback {
	memset((void *)&audioFormat, 0, sizeof(audioFormat)); // recommended

	Float64 hardwareSampleRate;
	UInt32 propertySize = sizeof(hardwareSampleRate);	
	AudioSessionGetProperty ( kAudioSessionProperty_CurrentHardwareSampleRate,
		&propertySize, &hardwareSampleRate);

	if (audioInputFlag == 1)
		audioFormat.mSampleRate = hardwareSampleRate; // we have to use hardware rate for input
	else
		audioFormat.mSampleRate = 44100.0;

	audioFormat.mFormatID = kAudioFormatLinearPCM;
	audioFormat.mFormatFlags = kLinearPCMFormatFlagIsSignedInteger | kLinearPCMFormatFlagIsPacked;
	audioFormat.mBitsPerChannel = 16;
	audioFormat.mChannelsPerFrame = numberOfChannels;
	audioFormat.mBytesPerFrame = numberOfChannels*2; // assumes 16-bit samples, 2 bytes
	audioFormat.mFramesPerPacket = 1;
	audioFormat.mBytesPerPacket = audioFormat.mBytesPerFrame * audioFormat.mFramesPerPacket;
	audioFormat.mReserved = 0;
}

- (int) setupRemoteIO
{	
	int outputBus = 0;
	int inputBus = 1;

	// Open the output unit
	AudioComponentDescription desc;
	desc.componentType = kAudioUnitType_Output;
	desc.componentSubType = kAudioUnitSubType_RemoteIO;
	desc.componentManufacturer = kAudioUnitManufacturer_Apple;
	desc.componentFlags = 0;
	desc.componentFlagsMask = 0;

	AudioComponent comp = AudioComponentFindNext(NULL, &desc);
	AudioComponentInstanceNew(comp, &rioUnit);

	UInt32 one = 1;
	AudioUnitSetProperty(rioUnit, kAudioOutputUnitProperty_EnableIO,
		kAudioUnitScope_Output, outputBus, &one, sizeof(one));
	AudioUnitSetProperty(rioUnit, kAudioUnitProperty_StreamFormat,
		kAudioUnitScope_Input, outputBus, &audioFormat, sizeof(audioFormat));
	AudioUnitSetProperty(rioUnit, kAudioUnitProperty_SetRenderCallback,
		kAudioUnitScope_Global, outputBus, &inputProc, sizeof(inputProc));

	if (audioInputFlag == 1) // enable input side
		AudioUnitSetProperty(rioUnit, kAudioOutputUnitProperty_EnableIO,
			kAudioUnitScope_Input, inputBus, &one, sizeof(one));

	AudioUnitInitialize(rioUnit);

	return 0;
}

- (void) setupPlaybackAudioQueueObject {
	// create the playback audio queue object
	AudioQueueNewOutput (
		&audioFormat,
		playbackCallback,
		self, 
		CFRunLoopGetCurrent (),
		kCFRunLoopCommonModes,
		0,								// run loop flags
		&queueObject
	);
	
	// set the volume of the playback audio queue
	AudioQueueSetParameter (
		queueObject,
		kAudioQueueParam_Volume,
		1.0
	);
}

- (void) setupAudioQueueBuffers {
	// set basic queue/callback params
	//[self setBufferByteSize: vectorSize * audioFormat.mBytesPerFrame];
	bufferByteSize = vectorSize * audioFormat.mBytesPerFrame;

	// set rtcmix params (note we don't use inbuf and outbuf, they are passed in pulltraverse() above)
	maxmsp_rtsetparams(audioFormat.mSampleRate, audioFormat.mChannelsPerFrame, vectorSize, NULL, NULL, errbuf);

	// prime the queue with some data before starting
	// allocate and enqueue buffers				
	int bufferIndex;
	for (bufferIndex = 0; bufferIndex < numberAudioDataBuffers; ++bufferIndex) {
		AudioQueueAllocateBuffer (
			[self queueObject],
			[self bufferByteSize],
			&buffers[bufferIndex]
		);
		playbackCallback ( 
			self,
			[self queueObject],
			buffers[bufferIndex]
		);
	}
}

- (void) playAudio {
	if (audioUnitFlag == 0) 
	{
		[self setupAudioQueueBuffers];
		AudioQueueStart (self.queueObject, NULL);
	}
	else 
	{
		AudioOutputUnitStart([self rioUnit]);
	}
}

- (void) stopAudio {
	if (audioUnitFlag == 0) 
	{
		AudioQueueStop (self.queueObject, self.audioPlayerShouldStopImmediately);
	} 
	else 
	{
		AudioOutputUnitStop([self rioUnit]);
	}
}


- (void) pauseAudio {
	if (audioUnitFlag == 0) 
	{
		AudioQueuePause (self.queueObject);
	} 
	else 
	{
		AudioOutputUnitStop([self rioUnit]);
	}
}

- (void) resumeAudio {
	UInt32 sessionCategory;
	sessionCategory = kAudioSessionCategory_MediaPlayback;
	if (audioInputFlag == 1) 
		sessionCategory = kAudioSessionCategory_PlayAndRecord;
	
	AudioSessionSetProperty (kAudioSessionProperty_AudioCategory, sizeof (sessionCategory), &sessionCategory);
	AudioSessionSetActive (true);
	
	if (audioUnitFlag == 0) 
	{
		AudioQueueStart (self.queueObject, NULL);
	} 
	else 
	{
		AudioOutputUnitStart([self rioUnit]);
	}
}

- (void) disposeAudioQueue {
	[self flushAllScripts];
	[self setAudioPlayerShouldStopImmediately: YES];	
	[self stopAudio];
	AudioQueueDispose(queueObject,YES);
}

- (void) restartAudioQueue {
	[self setupPlayback];
	[self setupPlaybackAudioQueueObject];
	[self setAudioPlayerShouldStopImmediately: NO];	
	[self playAudio];
}

#pragma mark Singleton Methods

+ (id)sharedManager {
    @synchronized(self)
	{
        if (sharedSingletonManager == nil) 
		{
            [[self alloc] init]; // assignment not done here
        }
    }
    return sharedSingletonManager;
}

+ (id)allocWithZone:(NSZone *)zone
{
    @synchronized(self) 
	{
        if (sharedSingletonManager == nil)
		{
            sharedSingletonManager = [super allocWithZone:zone];
            return sharedSingletonManager;  // assignment and return on first allocation
        }
    }
    return nil; //on subsequent allocation attempts return nil
}

- (id)copyWithZone:(NSZone *)zone
{
    return self;
}

- (id)retain
{
    return self;
}

- (unsigned)retainCount
{
    return UINT_MAX;  //denotes an object that cannot be released
}

- (void)release
{
    //do nothing
}

- (id)autorelease
{
    return self;
}

- (void) dealloc {
	[self disposeAudioQueue];
	
	[super dealloc];
}

#pragma mark Playback and Interrupt Callbacks

// This is the callback function that does the sample-crunching work
// the AudioQueueBuffers get filled and sent off from here
static void playbackCallback (
							  void					*inUserData,
							  AudioQueueRef			inAudioQueue,
							  AudioQueueBufferRef		bufferReference
							  ) {
	// This callback, being 'outside' the implementation, needs references to the
	// RTcmixPlayer and RTcmixManager objects
	RTcmixPlayer *player = (RTcmixPlayer *) inUserData;
	//RTcmixManager *rtcmixManager = [RTcmixManager sharedManager];
	
	UInt32 numBytes = [player bufferByteSize];
	UInt32 numPackets = vectorSize;
	
	bufferReference->mAudioDataByteSize			= numBytes;		
	bufferReference->mPacketDescriptionCount	= numPackets;
	
	// BGG -- this drives the RTcmix sample-computation buffer
	short *bptr = bufferReference->mAudioData; // fill the AudioQueueBuffer directly in RTcmix
	pullTraverse(NULL, bptr); // no input audio, use the AudioUnit version
	
	if (check_bang() == 1) // check for retriggering, set by "MAXBANG" in the score
	{ 
		[player maxBang];
	}
	
	AudioQueueEnqueueBuffer (
							 inAudioQueue,
							 bufferReference,
							 0,
							 NULL
							 );
}

// this callback is used by RemoteIO AudioUnit if AudioUnit is employed instead of AudioQueueBuffers
static OSStatus	rtcmixPerformCallback (
									   void						*inRefCon, 
									   AudioUnitRenderActionFlags 	*ioActionFlags, 
									   const AudioTimeStamp 		*inTimeStamp, 
									   UInt32 						inBusNumber, 
									   UInt32 						inNumberFrames, 
									   AudioBufferList 			*ioData)
{
	RTcmixPlayer *player = (RTcmixPlayer *) inRefCon;
	//RTcmixManager *rtcmixManager = [RTcmixManager sharedManager];
	
	AudioUnitRender([player rioUnit], ioActionFlags, inTimeStamp, 1, inNumberFrames, ioData);
	
	// these buffers are used because RTcmix is working asynch with the AudioUnit
	short *rtcmixoutptr = player.rtcmixOutputBuffer; // pointers to the bufs we use
	short *rtcmixinptr = NULL;
	if (player.audioInputFlag == 1) 
		rtcmixinptr = player.rtcmixInputBuffer;
	
	//loop through all the buffers that need to be filled
	for (int i = 0 ; i < ioData->mNumberBuffers; i++){
		//get the buffer to be filled
		AudioBuffer buffer = ioData->mBuffers[i];
		UInt16 *frameBuffer = buffer.mData;
		
		//loop through the buffer and fill the frames
		for (int j = 0; j < inNumberFrames*numberOfChannels; j += numberOfChannels){
			for (int k = 0; k < numberOfChannels; k++) {
				if(player.outputBufferIndex >= vectorSize*numberOfChannels) 
				{
					pullTraverse(rtcmixinptr, rtcmixoutptr);
					if (check_bang() == 1) // check for retriggering, set by "MAXBANG" in the score
					{ 
						[player maxBang];
					}
					
					maxmessage_nvals = check_vals(maxmessage_vals); // check for messages, set by "MAXMESSAGE" in the score
					if (maxmessage_nvals)
					{
						[player maxMessage];
					}
					
					player.outputBufferIndex = 0;
					player.inputBufferIndex = 0;
				}
				if (player.audioInputFlag == 1)
					rtcmixinptr[player.inputBufferIndex+k] = frameBuffer[j+k];
				
				frameBuffer[j+k] = rtcmixoutptr[player.outputBufferIndex+k];
			}
			player.outputBufferIndex += numberOfChannels;
			player.inputBufferIndex += numberOfChannels;
		}
	}
	
	// used to return an OSStatus err from AudioUnitRender but it wouldna go laddie!
    return noErr;
}

void interruptionListenerCallback (
								   void	*inUserData,
								   UInt32	interruptionState
								   ) {
	RTcmixPlayer *player = (RTcmixPlayer *) inUserData;	
	
	if (interruptionState == kAudioSessionBeginInterruption) 
	{
		[player pauseAudio];
	}
	else if (interruptionState == kAudioSessionEndInterruption) 
	{
		[player resumeAudio];
	}
}

void propListenerCallback(
						  void *					inClientData,
						  AudioSessionPropertyID	inID,
						  UInt32                  inDataSize,
						  const void *            inData)
{
	RTcmixPlayer *player = (RTcmixPlayer *) inClientData;
	if (inID == kAudioSessionProperty_AudioRouteChange)
	{
		// if there was a route change, we need to dispose the current rio unit and create a new one
		AudioComponentInstanceDispose([player rioUnit]);		
		
		[player setupPlayback];
		[player setupRemoteIO];
		
		AudioOutputUnitStart([player rioUnit]);
	}
}

@end
